/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache;

import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.cbs.dao.LabResultRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.LabResultRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.VaccinationRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.Mapper;
import dk.vaccinationsregister.schemas._2021._04._01.Covid19VaccinationCardType;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.lang.NonNull;

import javax.sql.DataSource;
import java.io.IOException;
import java.io.Reader;
import java.util.function.Function;
import java.util.function.Supplier;

import static dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordMappers.mapper;
import static dk.sundhedsdatastyrelsen.cbs.datacache.impl.Writers.writer;

@Configuration
public class DataCacheImplConfiguration {

    @Bean
    @NonNull
    public ObjectMapper objectMapper() {
        return new ObjectMapper().findAndRegisterModules();
    }

    @Bean
    @NonNull
    public Writer<VaccinationRecord<Vaccination, Covid19VaccinationCardType>> vaccinationWriter(
        @NonNull final DataSource dataSource,
        @NonNull final Mapper<VaccinationRecordImpl<Vaccination, Covid19VaccinationCardType>> mapper) {
        return writer(dataSource, mapper);
    }

    @Bean
    @NonNull
    public Mapper<VaccinationRecordImpl<Vaccination, Covid19VaccinationCardType>> vaccinationMapper(
        @NonNull final Supplier<VaccinationRecordImpl<Vaccination, Covid19VaccinationCardType>> constructor,
        @NonNull final ObjectMapper mapper) {
        return mapper(
            VaccinationRecord.CATALOG_NAME,
            VaccinationRecord.SCHEMA_NAME,
            VaccinationRecord.TABLE_NAME,
            constructor,
            deserializer(mapper, Vaccination.class),
            serializer(mapper),
            deserializer(mapper, Covid19VaccinationCardType.class),
            serializer(mapper))
            .build();
    }

    @Bean
    @NonNull
    public Writer<LabResultRecord<LabResult, Covid19VaccinationCardType>> labResultWriter(
        @NonNull final DataSource dataSource,
        @NonNull final Mapper<LabResultRecordImpl<LabResult, Covid19VaccinationCardType>> mapper) {
        return writer(dataSource, mapper);
    }

    @Bean
    @NonNull
    public Mapper<LabResultRecordImpl<LabResult, Covid19VaccinationCardType>> labResultMapper(
        @NonNull final Supplier<LabResultRecordImpl<LabResult, Covid19VaccinationCardType>> constructor,
        @NonNull final ObjectMapper mapper) {
        return mapper(
            LabResultRecord.CATALOG_NAME,
            LabResultRecord.SCHEMA_NAME,
            LabResultRecord.TABLE_NAME,
            constructor,
            deserializer(mapper, LabResult.class),
            serializer(mapper),
            deserializer(mapper, Covid19VaccinationCardType.class),
            serializer(mapper))
            .build();
    }

    @Bean
    @NonNull
    public Supplier<VaccinationRecordImpl<Vaccination, Covid19VaccinationCardType>> vaccinationConstructor() {
        return VaccinationRecordImpl::new;
    }

    @Bean
    @NonNull
    public Supplier<LabResultRecordImpl<LabResult, Covid19VaccinationCardType>> labResultConstructor() {
        return LabResultRecordImpl::new;
    }

    private static <T> Function<Reader, T> deserializer(@NonNull final ObjectMapper mapper, @NonNull final Class<T> clazz) {
        return reader -> {
            try {
                return mapper.readValue(reader, clazz);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        };
    }

    private static <T> Function<T, String> serializer(@NonNull final ObjectMapper mapper) {
        return object -> {
            try {
                return mapper.writeValueAsString(object);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        };
    }

}
