/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.liquibase;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import javax.sql.DataSource;
import liquibase.Contexts;
import liquibase.LabelExpression;
import liquibase.Liquibase;
import liquibase.database.DatabaseFactory;
import liquibase.database.jvm.JdbcConnection;
import liquibase.exception.LiquibaseException;
import liquibase.resource.ClassLoaderResourceAccessor;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.task.SimpleAsyncTaskExecutor;
import org.springframework.jdbc.datasource.DriverManagerDataSource;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;

@Service
public class AsyncLiquibaseService {

    @Value("${liquibase.changelog.file}")
    private String changeLogFile;

    @Value("${liquibase.connection.polling.delay}")
    private long pollingDelay;

    @Value("${liquibase.connection.polling.attempts}")
    private int pollingAttempts;

    @Autowired
    private final SimpleAsyncTaskExecutor executor;

    private final DataSource dataSource;
    private static final Logger log = Logger.getLogger(AsyncLiquibaseService.class);

    public AsyncLiquibaseService(
        @Value("${spring.datasource.driver}") String driverClassName,
        @Value("${spring.datasource.url}") String url,
        @Value("${liquibase.datasource.username}") String username,
        @Value("${liquibase.datasource.password}") String password,
        SimpleAsyncTaskExecutor executor
    ) {
        DriverManagerDataSource ds = new DriverManagerDataSource();
        ds.setDriverClassName(driverClassName);
        ds.setUrl(url);
        ds.setUsername(username);
        ds.setPassword(password);
        this.dataSource = ds;

        this.executor = executor;
    }

    /**
     * Configures Liquibase so that it can be triggered at runtime after the database is up and running.
     * The DATABASECHANGELOG entry will be tagged with the project version for convenient rollback.
     * <p>
     * The bean polls the database until a connection is established.  This is to ensure the database is up and running
     * before Liquibase attempts to update
     */
    @Async
    public void update() {
        try (Connection connection = dataSource.getConnection()) {
            boolean isConnectionAvailable = pollForDatabaseConnection(connection);

            if (isConnectionAvailable) {
                log.info("Updating cbs-datacache database from: " + changeLogFile + "'");

                try (Liquibase liquibase = new Liquibase(
                    changeLogFile,
                    new ClassLoaderResourceAccessor(),
                    DatabaseFactory.getInstance().findCorrectDatabaseImplementation(new JdbcConnection(connection))
                )) {
                    liquibase.update(new Contexts(), new LabelExpression());
                }
            } else {
                throw new SQLException();
            }
        } catch (SQLException e) {
            log.warn("Database connection failed after " + pollingAttempts + " attempts.  " +
                "Liquibase change sets may not have been applied: " + e.getMessage());
        } catch (LiquibaseException e) {
            log.warn("An exception occurred. Liquibase change sets may not have been applied correctly: " + e.getMessage());
        }
    }

    /**
     * Polls the database connection using a synchronised lock
     *
     * @param connection database connection to use
     * @return true on successful connection or false if number of tries has been exceeded
     */
    private boolean pollForDatabaseConnection(Connection connection) {
        boolean isConnectionAvailable = false;
        int attemptsRemaining = pollingAttempts;

        synchronized (executor) {
            try {
                while (!isConnectionAvailable && attemptsRemaining > 0) {
                    log.info("Polling database connection after " + pollingDelay + " ms.  "
                        + attemptsRemaining + " attempts remaining");
                    executor.wait(pollingDelay);
                    isConnectionAvailable = isDatabaseAlive(connection);
                    attemptsRemaining--;
                }
            } catch (InterruptedException e) {
                log.warn(e.getMessage());
                Thread.currentThread().interrupt();
            }
        }
        return isConnectionAvailable;
    }

    /**
     * Attempt to send a query to the database
     *
     * @param connection to database
     * @return true if select query returns something
     */
    private boolean isDatabaseAlive(Connection connection) {
        ResultSet result = null;

        try {
            if (connection == null || connection.isClosed()) {
                return false;
            }

            try (Statement statement = connection.createStatement()) {
                result = statement.executeQuery("select 1");
            }

        } catch (SQLException e) {
            log.debug(e.getMessage());
        }

        return result != null;
    }
}
