/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.config;

import com.sun.org.apache.xerces.internal.jaxp.datatype.XMLGregorianCalendarImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.VaccinationRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.LabResultETL;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.VaccinationETL;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.vaccinationsregister.schemas._2021._04._01.Covid19VaccinationCardType;
import dk.vaccinationsregister.schemas._2021._04._01.EffectuatedItemType;
import dk.vaccinationsregister.schemas._2021._04._01.VaccinationPlanType;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.convert.converter.Converter;
import org.springframework.lang.NonNull;

import java.time.LocalDateTime;
import java.util.function.BiFunction;
import java.util.function.Function;

@Configuration
public class DataCacheConfiguration {

    /**
     * Produces an ETL (extract transform load) for vaccinations
     *
     * @param extractor         the extractor, extracts an DDV dto using a cpr number
     * @param dtoTransformer    the dto transformer, transforms the DDV dto to the internal dto
     * @param dtoDaoTransformer the dto to dao transformer, transforms the DDV dto and the internal dto to the dao
     * @param writer            the writer, loads the dao return <code>true</code> when load succeeded else <code>false</code>
     * @return the ETL converter
     */
    @Bean
    @NonNull
    @VaccinationETL
    public Function<String, Boolean> vaccination(
        @NonNull final Function<String, Covid19VaccinationCardType> extractor,
        @NonNull final Converter<Covid19VaccinationCardType, Vaccination> dtoTransformer,
        @NonNull final BiFunction<Vaccination, Covid19VaccinationCardType, VaccinationRecord<Vaccination, Covid19VaccinationCardType>> dtoDaoTransformer,
        @NonNull final Writer<VaccinationRecord<Vaccination, Covid19VaccinationCardType>> writer) {
        return cpr -> {
            final Covid19VaccinationCardType card = extractor.apply(cpr);
            return writer.write(dtoDaoTransformer.apply(dtoTransformer.convert(card), card));
        };
    }

    @Bean
    @NonNull
    public BiFunction<Vaccination, Covid19VaccinationCardType, VaccinationRecord<Vaccination, Covid19VaccinationCardType>> vaccinationRecordConverter() {
        return (data, source) -> new VaccinationRecordImpl<Vaccination, Covid19VaccinationCardType>()
            .cpr(data.cpr())
            .creationTime(LocalDateTime.now())
            .traceId("TODO")
            .setData(data)
            .setSource(source);
    }

    /**
     * Mock Extractor for DDV
     */
    @Bean
    @NonNull
    public Function<String, Covid19VaccinationCardType> mockDdvExtractor() {
        return cpr -> {
            Covid19VaccinationCardType cardType = new Covid19VaccinationCardType();
            cardType.setPersonCivilRegistrationIdentifier(cpr);

            EffectuatedItemType effectuatedItem = new EffectuatedItemType();
            effectuatedItem.setEffectuatedVaccineName("My vaccine");
            effectuatedItem.setEffectuatedVaccinationDateTime(XMLGregorianCalendarImpl.createDateTime(2021, 4, 14, 15, 0, 0));
            cardType.getEffectuatedItem().add(effectuatedItem);

            VaccinationPlanType vaccinationPlan = new VaccinationPlanType();
            vaccinationPlan.setVaccinationPlanName("My vaccination plan");
            cardType.getVaccinationPlan().add(vaccinationPlan);
            return cardType;
        };
    }

    /**
     * Produces an ETL (extract transform load) for lab results
     *
     * @param extractor         the extractor, extracts an KIDS dto using a cpr number
     * @param dtoTransformer    the transformer, transforms the KIDS dto to the internal dto
     * @param dtoDaoTransformer the dto to dao transformer, transforms the KIDS dto and the internal dto to the dao
     * @param writer            the writer, loads the dao return <code>true</code> when load succeeded else <code>false</code>
     * @return the ETL converter
     */
    @Bean
    @NonNull
    @LabResultETL
    public Function<String, Boolean> labResult()
//    @NonNull final Function<String, Covid19VaccinationCardType> extractor,
//    @NonNull final Converter<Covid19VaccinationCardType, LabResult> dtoTransformer,
//    @NonNull final BiFunction<LabResult, Covid19VaccinationCardType, LabResultRecord<LabResult, Covid19VaccinationCardType>> dtoDaoTransformer,
//    @NonNull final Writer<LabResultRecord<LabResult, Covid19VaccinationCardType>> writer)
    {
        return cpr -> true;
    }
}
