/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.ddv;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.sdsd.dgws._2012._06.WhiteListingHeader;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientService;
import dk.vaccinationsregister.schemas._2021._04._01.GetCovid19VaccinationCardRequestType;
import dk.vaccinationsregister.schemas._2021._04._01.GetCovid19VaccinationCardResponseType;
import dk.vaccinationsregister.schemas._2021._04._01.ObjectFactory;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.xmlsoap.schemas.soap.envelope.Body;
import org.xmlsoap.schemas.soap.envelope.Envelope;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import java.io.StringReader;
import java.util.List;

public class DDVWebService {

    private static final org.xmlsoap.schemas.soap.envelope.ObjectFactory SOAP_FACTORY = new org.xmlsoap.schemas.soap.envelope.ObjectFactory();

    // Quotes are deliberate (see https://www.w3.org/TR/2000/NOTE-SOAP-20000508/#_Toc478383528)
    private static final String DDV_SOAP_ACTION = "\"http://vaccinationsregister.dk/schemas/2021/04/01#GetCovid19VaccinationCard\"";

    private final String serviceEndpoint;
    private final DGWSHelper dgwsHelper;
    private final JAXBContext jaxbContext;
    private final HttpClientService httpClientService;

    public DDVWebService(DGWSHelper dgwsHelper, String serviceEndpoint, HttpClientService httpClientService) {
        this.serviceEndpoint = serviceEndpoint;
        this.dgwsHelper = dgwsHelper;
        this.httpClientService = httpClientService;
        try {
            this.jaxbContext = JAXBContext.newInstance(Envelope.class, Security.class, Header.class, WhiteListingHeader.class, GetCovid19VaccinationCardRequestType.class, GetCovid19VaccinationCardResponseType.class);
        } catch (JAXBException e) {
            throw new IllegalStateException(e);
        }
    }

    public GetCovid19VaccinationCardResponseType getCovid19VaccinationCard(String cpr) throws JAXBException {
        GetCovid19VaccinationCardRequestType request = new GetCovid19VaccinationCardRequestType();
        request.setPersonCivilRegistrationIdentifier(cpr);
        dgwsHelper.ensureValidIdCard();
        return callService(
            DDV_SOAP_ACTION,
            new ObjectFactory().createGetCovid19VaccinationCardRequest(request),
            serviceEndpoint);
    }

    private <RequestType, ResponseType> ResponseType callService(String soapAction, RequestType request, String endpoint) throws JAXBException {
        Body soapBody = SOAP_FACTORY.createBody();
        soapBody.getAny().add(request);
        org.xmlsoap.schemas.soap.envelope.Header soapHeader = SOAP_FACTORY.createHeader();
        Envelope soapEnvelope = SOAP_FACTORY.createEnvelope();
        String newRequest = dgwsHelper.addHeaders(jaxbContext, soapBody);
        soapEnvelope.setHeader(soapHeader);
        soapEnvelope.setBody(soapBody);

        String responseXml = httpClientService.soapRequest(endpoint, soapAction, newRequest);
        Envelope responseEnvelope = (Envelope) jaxbContext.createUnmarshaller().unmarshal(new StringReader(responseXml));

        try {
            // Body has type "any", so we need additional unmarshalling
            List<Object> any = responseEnvelope.getBody().getAny();
            JAXBElement unmarshalled = (JAXBElement) JAXBContext.newInstance(ObjectFactory.class).createUnmarshaller().unmarshal((org.w3c.dom.Node) any.get(0));
            return (ResponseType) unmarshalled.getValue();
        } catch (Exception e) {
            throw new RuntimeException("Unable to parse body", e);
        }
    }

}
