/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import COSE.AlgorithmID;
import COSE.CoseException;
import COSE.OneKey;
import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.cbs.service.webservice.hcertmock.*;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.*;
import org.springframework.stereotype.Service;

import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

@Service
public class MockDataService {

    static OneKey cborPrivateKey;

    public MockDataService() {
        try {
            cborPrivateKey = OneKey.generateKey(AlgorithmID.ECDSA_256);
        } catch (CoseException e) {
            throw new IllegalStateException(e);
        }
    }

    private static Map<String, String> jwts = new HashMap<>();

    static {
        jwts.put("DK", "");
        jwts.put("DKPLUS", "");
        jwts.put("EU", "");
    }

    public GetCoronaPassResponse cacheMiss(String jobId) {
        GetCoronaPassResponse getCoronaPassResponse = new GetCoronaPassResponse();
        FutureResult futureResult = new FutureResult();
        futureResult.setJobId(jobId);
        getCoronaPassResponse.setFutureResult(futureResult);

        return getCoronaPassResponse;
    }

    public GetCoronaPassResponse cacheHit(String cpr) {
        try {
            GetCoronaPassResponse getCoronaPassResponse = new GetCoronaPassResponse();
            getCoronaPassResponse.setCoronaPasses(new PassList());

            if (emptyAnswerForCPR(cpr)) return getCoronaPassResponse;

            Person person = person(cpr);
            for (String type : jwts.keySet()) {
                PassJWT passJWT = new PassJWT();
                passJWT.setType(type);

                if ("EU".equals(type) && person.euPass) {
                    CertificatePayload cp = getEUCertificatePayload(cpr);
                    if (cp != null) {
                        String json = new ObjectMapper().writeValueAsString(cp);
                        String encoded = new GreenCertificateEncoder(cborPrivateKey, UUID.randomUUID().toString(), "HC1").encode(json);
                        passJWT.setJWT(encoded == null ? "" : encoded);

                        KeyValuePair kv = new KeyValuePair();
                        kv.setKey("eu:keytype");
                        kv.setValue("digital_green_certificate");
                        passJWT.getProperties().add(kv);
                    }
                }
                if ("DK".equals(type) && person.dkPass) {
                    String json = new ObjectMapper().writeValueAsString(new DKDSL().withValidCertificate(cpr));
                    String encoded = new GreenCertificateEncoder(cborPrivateKey, UUID.randomUUID().toString(), "DK1").encode(json);
                    passJWT.setJWT(encoded == null ? "" : encoded);
                }
                if ("DKPLUS".equals(type) && person.dkPass) {
                    String json = new ObjectMapper().writeValueAsString(new DKPlusDSL().withValidCertificate(cpr, person.first + " " + person.last,"1987-05-04"));
                    String encoded = new GreenCertificateEncoder(cborPrivateKey, UUID.randomUUID().toString(), "DK2").encode(json);
                    passJWT.setJWT(encoded == null ? "" : encoded);
                }
                if (passJWT.getJWT() != null && !"".equals(passJWT.getJWT())) {
                    getCoronaPassResponse.getCoronaPasses().getPass().add(passJWT);
                }
            }
            return getCoronaPassResponse;
        } catch (Exception e) {
            throw new IllegalStateException(e);
        }
    }

    private boolean emptyAnswerForCPR(String cpr) {
        if ("0405870114".equals(cpr) || person(cpr) == null) {
            return true;
        } else {
            return false;
        }
    }

    private static class Person {
        String first;
        String last;
        boolean euPass;
        boolean dkPass;
        public Person(String first, String last, boolean euPass, boolean dkPass) {
            this.first = first;
            this.last = last;
            this.euPass = euPass;
            this.dkPass = dkPass;
        }
    }

    private Person person(String cpr) {
        switch (cpr) {
            case "0405870101":
                return new Person("Anders", "Andersen", true, true);
            case "0405870102":
                return new Person("Bente", "Bendtsen", true, true);
            case "0405870103":
                return new Person("Casper", "Christensen", true, true);
            case "0405870104":
                return new Person("Dorthe", "Drewes", true, true);
            case "0405870105":
                return new Person("Egon", "Eilersen", true, true);
            case "0405870106":
                return new Person("Frederikke", "Frederiksen", true, true);
            case "0405870107":
                return new Person("Gert", "Gorbatjov", true, true);
            case "0405870108":
                return new Person("Helle", "Henriksen", true, true);
            case "0405870109":
                return new Person("Ivan", "Ingolfsen", true, true);
            case "0405870110":
                return new Person("Jytte", "Jensen", true, true);
            case "0405870111":
                return new Person("Klaus", "Kornsted", false, true);
            case "0405870112":
                return new Person("Line", "Larsen", false, true);
            case "0405870113":
                return new Person("Martin", "Monradsen", false, true);
            case "0405870115":
                return new Person("Olga", "Olsen", true, false);
            case "0405870117":
                return new Person("Quang", "Que", true, true);
            case "0405870118":
                return new Person("Rasmus", "Rossen", true, true);
            default:
                return null;
        }
    }

    public CertificatePayload getEUCertificatePayload(String cpr) {
        Person name = person(cpr);
        switch (cpr) {
            case "0405870101":
            case "0405870102":
            case "0405870115":
            case "0405870117":
            case "0405870118":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                        .withVaccine()
                        .withTestResult()
                        .withRecovery()
                        .build(cpr);
            case "0405870103":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                        .withVaccine()
                        .withTestResult()
                        .withExpiredRecovery()
                        .build(cpr);
            case "0405870104":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                        .withVaccine()
                        .withExpiredTestResult()
                        .withExpiredRecovery()
                        .build(cpr);
            case "0405870105":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                        .withExpiredVaccine()
                        .withExpiredTestResult()
                        .withExpiredRecovery()
                        .build(cpr);
            case "0405870106":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                        .withExpiredVaccine()
                        .withTestResult()
                        .withExpiredRecovery()
                        .build(cpr);
            case "0405870107":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                        .withExpiredVaccine()
                        .withTestResult()
                        .withRecovery()
                        .build(cpr);
            case "0405870108":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                        .withVaccine()
                        .withExpiredTestResult()
                        .withRecovery()
                        .build(cpr);
            case "0405870109":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                        .withExpiredVaccine()
                        .withExpiredTestResult()
                        .withRecovery()
                        .build(cpr);
            case "0405870110":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                        .withVaccine()
                        .withTestResult()
                        .withRecovery()
                        .expiredBuild();
            case "0405870111":
            case "0405870112":
            case "0405870113":
                return null;
            default:
                return null;
        }
    }

}
