/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import COSE.AlgorithmID;
import COSE.OneKey;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import org.junit.jupiter.api.Test;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Date;
import java.util.TimeZone;

import static org.junit.jupiter.api.Assertions.*;

@CbsTest
class CoronaPassWebServiceTest {
    private final ApplicationContext applicationContext = new AnnotationConfigApplicationContext("dk.sundhedsdatastyrelsen.cbs.test");
    private final GetCoronaPassRequest request = new GetCoronaPassRequest();
    private final CoronaPassWebService coronaPassWebService = new CoronaPassWebService(applicationContext);

    private GetCoronaPassRequest request(String cpr) {
        GetCoronaPassRequest request = new GetCoronaPassRequest();
        request.setIdentifier(cpr);
        return request;
    }

    @Test
    void test_getCoronaPass_cache_hit() {
        System.setProperty("test.cache.hit", "true");
        String cprPrefix = "040587010";
        for (int i=0; i<10; i++)
            verifyPass(cprPrefix + i);
    }

    @Test
    void test_getCoronaPass_empty() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870114";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(0, coronaPass.getCoronaPasses().getPass().size());
    }

    @Test
    void test_getCoronaPassOnlyDK() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870113";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(2, coronaPass.getCoronaPasses().getPass().size());
    }

    @Test
    void test_getCoronaPassOnlyEU() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870115";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(1, coronaPass.getCoronaPasses().getPass().size());
    }

    @Test
    void test_getCoronaPass5minutesExpiry() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870118";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(3, coronaPass.getCoronaPasses().getPass().size());
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        LocalDateTime expDate = getEUExp(eu);
        LocalDateTime now = LocalDateTime.now(TimeZone.getDefault().toZoneId());
        assertTrue(now.plusMinutes(4).isBefore(expDate));
        assertTrue(now.plusMinutes(6).isAfter(expDate));
    }

    @Test
    void test_getCoronaPass15minutesExpiry() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870117";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(3, coronaPass.getCoronaPasses().getPass().size());
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        LocalDateTime expDate = getEUExp(eu);
        LocalDateTime now = LocalDateTime.now(TimeZone.getDefault().toZoneId());
        assertTrue(now.plusMinutes(14).isBefore(expDate));
        assertTrue(now.plusMinutes(16).isAfter(expDate));
    }

    @Test
    void test_getCoronaPass72hourExpiry() {
        System.setProperty("test.cache.hit", "true");
        String cpr = "0405870107";
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(3, coronaPass.getCoronaPasses().getPass().size());
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        LocalDateTime expDate = getEUExp(eu);
        LocalDateTime now = LocalDateTime.now(TimeZone.getDefault().toZoneId());
        assertTrue(now.plusHours(71).isBefore(expDate));
        assertTrue(now.plusHours(75).isAfter(expDate));
    }

    private LocalDateTime getEUExp(PassJWT eu) {
        try {
            String decoded = new GreenCertificateDecoder(OneKey.generateKey(AlgorithmID.RSA_PSS_256)).decode(eu.getJWT());
            JsonNode jsonNode = new ObjectMapper().readTree(decoded);
            long exp = jsonNode.findPath("4").asLong();
            return LocalDateTime.ofInstant(Instant.ofEpochMilli(exp * 1000), ZoneId.of("UTC"));
        } catch (Exception e) {
            throw new IllegalStateException(e);
        }
    }

    private void verifyPass(String cpr) {
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
    }

    @Test
    void test_getCoronaPass_cache_miss() {
        System.setProperty("test.cache.hit", "false");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request);
        assertNotNull(coronaPass);
        assertNotNull(coronaPass.getFutureResult());
        assertNotNull(coronaPass.getFutureResult().getJobId());

        assertNull(coronaPass.getCoronaPasses());
    }
}

