/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import COSE.*;
import com.upokecenter.cbor.CBORObject;
import nl.minvws.encoding.Base45;
import org.apache.commons.compress.compressors.CompressorException;
import org.apache.commons.compress.compressors.CompressorInputStream;
import org.apache.commons.compress.compressors.CompressorStreamFactory;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

import static org.apache.commons.compress.utils.IOUtils.toByteArray;

public class GreenCertificateDecoder {

  private final OneKey publicKey;

  public GreenCertificateDecoder(OneKey publicKey) {
    this.publicKey = publicKey;
  }



  /**
   * Decodes base45 encoded string -> Deflate -> COSE -> CBOR -> arbitrary Json String
   *
   * @param base45String
   * @return
   * @throws CompressorException
   * @throws IOException
   * @throws CoseException
   */
  public String decode(String base45String) throws CompressorException, IOException, CoseException {
    //if(!base45String.startsWith("HC1"))
      //throw new RuntimeException("Base45 string not valid according to specification");

    base45String = base45String.substring(3);
    byte[] decodedBytes = Base45.getDecoder().decode(base45String);

    byte[] coseBytes = getCoseBytes(decodedBytes);

    byte[] cborBytes = getCborBytes(coseBytes);

    return getJsonString(cborBytes);
  }

  private String getJsonString(byte[] cborBytes) throws IOException {
    CBORObject cborObject = CBORObject.DecodeFromBytes(cborBytes);
    ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
    cborObject.WriteJSONTo(byteArrayOutputStream);

    return byteArrayOutputStream.toString("UTF-8");
  }

  private byte[] getCborBytes(byte[] coseBytes) throws CoseException {
    Sign1Message msg = (Sign1Message) Message.DecodeFromBytes(coseBytes, MessageTag.Sign1);
    //if (!msg.validate(publicKey)) {
      //throw new RuntimeException("Could not verify COSE signature");
    //}

    return msg.GetContent();
  }

  private byte[] getCoseBytes(byte[] decodedBytes) throws CompressorException, IOException {

    CompressorInputStream compressedStream = new CompressorStreamFactory()
        .createCompressorInputStream(CompressorStreamFactory.DEFLATE,
            new ByteArrayInputStream(decodedBytes));

    return toByteArray(compressedStream);
  }
}
