/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.fluent.spring.jdbc;

import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.sql.ResultSet;
import java.sql.Timestamp;
import java.sql.Types;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.Supplier;

public final class Mappers {

    private Mappers() {}

    @NonNull
    public static <T> MapperBuilder<T> mapper(
        @Nullable final String catalogName,
        @Nullable final String schemaName,
        @NonNull final String tableName,
        @NonNull final Supplier<T> constructor) {
        return new MapperBuilder<>(catalogName, schemaName, tableName, constructor);
    }

    public static final class MapperBuilder<T> {

        @Nullable private final String catalogName;
        @Nullable private final String schemaName;
        @NonNull private final String tableName;
        @NonNull private final Supplier<T> constructor;
        @NonNull private final List<BiConsumer<T, ResultSet>> deserializers;
        @NonNull private final Map<String, Serializer<T, ?>> serializers;

        private MapperBuilder(
            @Nullable final String catalogName,
            @Nullable final String schemaName,
            @NonNull final String tableName,
            @NonNull final Supplier<T> constructor) {
            this.catalogName = catalogName;
            this.schemaName = schemaName;
            this.tableName = tableName;
            this.constructor = constructor;
            deserializers = new ArrayList<>();
            serializers = new HashMap<>();
        }

        @NonNull
        public <U, V> MapperBuilder<T> addMapping(
            @NonNull final String columnLabel,
            @NonNull final Deserializer<T, U> deserializer,
            @NonNull final Serializer<T, V> serializer) {
            deserializers.add(
                (obj, resultSet) -> deserializer.setter.accept(obj, deserializer.mapper.apply(resultSet, columnLabel)));
            serializers.put(columnLabel, serializer);
            return this;
        }

        @NonNull
        public MapperBuilder<T> addTimestampMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, Timestamp> setter,
            @NonNull final Function<T, Timestamp> getter) {
            addMapping(columnLabel, timestamp(setter), timestamp(getter));
            return this;
        }

        @NonNull
        public MapperBuilder<T> addTimestampToLocalDateTimeMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, LocalDateTime> setter,
            @NonNull final Function<T, LocalDateTime> getter) {
            addMapping(columnLabel, timestampToLocalDateTime(setter), localDateTime(getter));
            return this;
        }

        @NonNull
        public MapperBuilder<T> addStringMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, String> setter,
            @NonNull final Function<T, String> getter) {
            addMapping(columnLabel, string(setter), string(getter));
            return this;
        }

        @NonNull
        public MapperBuilder<T> addIntegerMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, Integer> setter,
            @NonNull final Function<T, Integer> getter) {
            addMapping(columnLabel, integer(setter), integer(getter));
            return this;
        }

        @NonNull
        public MapperBuilder<T> addBytesMapping(
            @NonNull final String columnLabel,
            @NonNull final BiConsumer<T, UUID> setter,
            @NonNull final Function<T, UUID> getter) {
            addMapping(columnLabel, bytes(setter), uuid(getter));
            return this;
        }

        @NonNull
        public Mapper<T> build() {
            return new MapperImpl<>(
                catalogName,
                schemaName,
                tableName,
                constructor,
                deserializers,
                serializers);
        }
    }

    private static final class MapperImpl<T> implements Mapper<T> {

        @Nullable private final String catologName;
        @Nullable private final String schemaName;
        @NonNull private final String tableName;
        @NonNull private final Supplier<T> constructor;
        @NonNull private final List<BiConsumer<T, ResultSet>> deserializers;
        @NonNull private final Map<String, Serializer<T, ?>> serializers;

        private MapperImpl(
            @Nullable final String catalogName,
            @Nullable final String schemaName,
            @NonNull final String tableName,
            @NonNull final Supplier<T> constructor,
            @NonNull final List<BiConsumer<T, ResultSet>> deserializers,
            @NonNull final Map<String, Serializer<T, ?>> serializers) {
            this.catologName = catalogName;
            this.schemaName = schemaName;
            this.tableName = tableName;
            this.constructor = constructor;
            this.deserializers = deserializers;
            this.serializers = serializers;
        }

        @Override
        @NonNull
        public T mapRow(@NonNull final ResultSet resultSet, final int rowNum) {
            final T obj = constructor.get();
            deserializers.forEach(fieldMapper -> fieldMapper.accept(obj, resultSet));
            return obj;
        }

        @Override
        @Nullable
        public String catalogName() {
            return catologName;
        }

        @Override
        @Nullable
        public String schemaName() {
            return schemaName;
        }

        @Override
        @NonNull
        public String tableName() {
            return tableName;
        }

        @Override
        @NonNull
        public SqlParameterSource parameterSource(@NonNull final T object) {
            final MapSqlParameterSource source = new MapSqlParameterSource();
            serializers.forEach(
                (columnLabel, serializer) ->
                    source.addValue(
                        columnLabel,
                        serializer.getter.apply(object),
                        serializer.sqlType));
            return source;
        }
    }

    public static <T, U> SerializerChain<T, U> serializer(@NonNull final Function<T, U> getter) {
        return new SerializerChain<>(getter);
    }

    public static final class SerializerChain<T, U> {
        @NonNull private final Function<? super T, U> getter;

        private SerializerChain(@NonNull final Function<? super T, U> getter) {
            this.getter = getter;
        }

        @NonNull
        public <V> SerializerChain<T, V> map(@NonNull final Function<? super U, V> mapper) {
            return new SerializerChain<>(getter.andThen(mapper));
        }

        @NonNull
        public <V> Serializer<T, V> get(@NonNull final Function<? super U, V> mapper, final int sqlType) {
            return new Serializer<>(getter.andThen(mapper), sqlType);
        }
    }

    public static final class Serializer<T, U> {
        private final int sqlType;
        @NonNull private final Function<? super T, U> getter;

        private Serializer(@NonNull final Function<? super T, U> getter, final int sqlType) {
            this.sqlType = sqlType;
            this.getter = getter;
        }
    }

    @NonNull
    private static <T> Serializer<T, String> string(@NonNull final Function<T, String> getter) {
        return new Serializer<>(getter, Types.VARCHAR);
    }

    @NonNull
    private static <T> Serializer<T, Timestamp> timestamp(@NonNull final Function<T, Timestamp> getter) {
        return new Serializer<>(getter, Types.TIMESTAMP);
    }

    @NonNull
    public static <T> Serializer<T, Timestamp> localDateTime(@NonNull final Function<T, LocalDateTime> getter) {
        return timestamp(owner -> {
            final LocalDateTime dateTime = getter.apply(owner);
            return dateTime == null ? null : Timestamp.valueOf(dateTime);
        });
    }

    @NonNull
    public static <T> Serializer<T, Integer> integer(@NonNull final Function<T, Integer> getter) {
        return new Serializer<>(getter, Types.INTEGER);
    }

    @NonNull
    public static <T> Serializer<T, byte[]> uuid(@NonNull final Function<T, UUID> getter) {
        return new Serializer<>(getter.andThen(UUIDs::toBytes), Types.BINARY);
    }

    @NonNull
    public static <T> DeserializerChain<T> deserializer(@NonNull final BiFunction<ResultSet, String, T> function) {
        return new DeserializerChain<>(function);
    }

    public static final class DeserializerChain<T> {

        @NonNull private final BiFunction<ResultSet, String, T> mapper;

        private DeserializerChain(@NonNull final BiFunction<ResultSet, String, T> mapper) {
            this.mapper = mapper;
        }

        @NonNull
        public <U> DeserializerChain<U> map(@NonNull final Function<T, U> mapper) {
            return new DeserializerChain<>(this.mapper.andThen(mapper));
        }

        @NonNull
        public <U> Deserializer<U, T> set(@NonNull final BiConsumer<U, ? super T> setter) {
            return new Deserializer<>(mapper, setter);
        }
    }

    public static final class Deserializer<T, U> {

        @NonNull private final BiFunction<ResultSet, String, U> mapper;
        @NonNull private final BiConsumer<T, ? super U> setter;

        private Deserializer(
            @NonNull final BiFunction<ResultSet, String, U> mapper,
            @NonNull final BiConsumer<T, ? super U> setter) {
            this.mapper = mapper;
            this.setter = setter;
        }
    }

    @NonNull
    public static <T> Deserializer<T, Timestamp> timestamp(@NonNull final BiConsumer<T, Timestamp> setter) {
        return deserializer(ResultSets::getTimestamp).set(setter);
    }

    @NonNull
    public static <T> Deserializer<T, LocalDateTime> timestampToLocalDateTime(@NonNull final BiConsumer<T, LocalDateTime> setter) {
        return deserializer(ResultSets::getTimestamp).map(Timestamp::toLocalDateTime).set(setter);
    }

    @NonNull
    public static <T> Deserializer<T, String> string(@NonNull final BiConsumer<T, String> setter) {
        return deserializer(ResultSets::getString).set(setter);
    }

    @NonNull
    public static <T> Deserializer<T, Integer> integer(@NonNull final BiConsumer<T, Integer> setter) {
        return deserializer(ResultSets::getInteger).set(setter);
    }

    @NonNull
    public static <T> Deserializer<T, UUID> bytes(@NonNull final BiConsumer<T, UUID> setter) {
        return deserializer(ResultSets::getBytes).map(UUIDs::of).set(setter);
    }
}
