/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.nas.facade;

import dk.nsi.advis.v10.NotifyContent;
import dk.sundhedsdatastyrelsen.cbs.dgws.DGWSRequestFactory;
import dk.sundhedsdatastyrelsen.cbs.dgws.STSConfig;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpExecutor;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.JobStatusNotification;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.ObjectFactory;
import org.apache.log4j.Logger;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.oasis_open.docs.wsn.b_2.DestroyPullPoint;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;
import org.oasis_open.docs.wsn.b_2.Unsubscribe;
import org.oasis_open.docs.wsn.bw_2.CreatePullPoint;
import org.oasis_open.docs.wsn.bw_2.NotificationBrokerService;
import org.oasis_open.docs.wsn.bw_2.NotificationConsumer;
import org.oasis_open.docs.wsn.bw_2.NotificationProducer;
import org.oasis_open.docs.wsn.bw_2.PullPointFactoryService;
import org.oasis_open.docs.wsn.bw_2.SubscriptionManagerService;

import javax.xml.bind.JAXBException;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Holder;
import javax.xml.ws.wsaddressing.W3CEndpointReference;
import java.util.List;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.assertEquals;

@CbsTest
@Disabled("Calls NAS")
class NasEndToEndIT {

    private static final Logger log = Logger.getLogger(NasEndToEndIT.class);

    public static final String NOTIFICATIONBROKER_SERVICE = "https://test1.ekstern-test.nspop.dk:8443/nas2/notificationbroker/service";
    public static final String PULLPOINTFACTORY_SERVICE = "https://test1.ekstern-test.nspop.dk:8443/nas2/pullpointfactory/service";
    public static final String SUBSCRIPTIONMANAGER_SERVICE = "https://test1.ekstern-test.nspop.dk:8443/nas2/subscriptionmanager/service";
    public static final String JOB_COMPLETED_TOPIC = "http://nspop.dk/GreenCertificate/2021/04/01:JobCompleted";

    private NasNotificationPublisher nasNotificationPublisher;
    private DGWSHeaderProvider headerProvider;
    private PullPointFactory pullPointFactory;
    private W3CEndpointReference pullPoint;
    private W3CEndpointReference subscribe;
    private NasClientImpl nasClient;

    @BeforeEach
    void setUp() throws JAXBException {
        STSConfig stsConfig = new STSConfig(
            "http://test1.ekstern-test.nspop.dk:8080/sts/services/NewSecurityTokenService",
            "Sundhedsdatastyrelsen",
            "33257872",
            "Sundhedsdatastyrelsen - COVID-19 Borgerstatus.jks",
            "Test1234",
            "CBS"
        );

        NotificationConsumer soap = new NotificationBrokerService().getSoap();
        setEndpoint((BindingProvider) soap, NOTIFICATIONBROKER_SERVICE);

        nasNotificationPublisher = new NasNotificationPublisher(soap, JOB_COMPLETED_TOPIC);
        headerProvider = new DefaultDGWSHeaderProvider(new DGWSRequestFactory(stsConfig, new dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientServiceImpl(new HttpExecutor())));
        pullPointFactory = new PullPointFactory();


        NotificationProducer notificationProducer = new SubscriptionManagerService().getSoap();
        setEndpoint((BindingProvider) notificationProducer, SUBSCRIPTIONMANAGER_SERVICE);

        CreatePullPoint createPullPoint = new PullPointFactoryService().getSoap();
        setEndpoint((BindingProvider) createPullPoint, PULLPOINTFACTORY_SERVICE);

        nasClient = new NasClientImpl(headerProvider, notificationProducer, createPullPoint, pullPointFactory);
    }

    @AfterEach
    void tearDown() throws Exception {
        String subscribeAddress = EndpointReferences.extractAddress(subscribe);
        NotificationProducer notificationProducer = new SubscriptionManagerService().getSoap();
        setEndpoint((BindingProvider) notificationProducer, subscribeAddress);
        DGWSHeaders headers = headerProvider.getHeaders();

        notificationProducer.unsubscribe(new Unsubscribe(), headers.getSecurity(), new Holder<>(headers.getMedComHeader()));
        pullPointFactory.getPullPoint(pullPoint).destroyPullPoint(new DestroyPullPoint(), headers.getSecurity(), new Holder<>(headers.getMedComHeader()));
    }

    @Test
    void subscribe_publish_and_get_messages() {
        pullPoint = nasClient.createPullPoint();
        subscribe = nasClient.subscribe(pullPoint, JOB_COMPLETED_TOPIC);

        JobStatusNotification jobStatusNotification = createJobStatusNotification();

        nasNotificationPublisher.publish(jobStatusNotification, "0506504003");

        List<NotificationMessageHolderType> messages = nasClient.getMessages(pullPoint, 1000, UUID.randomUUID().toString());

        Assertions.assertNotNull(messages);

        log.info(String.format("Found %s messages", messages.size()));

        messages.stream()
            .map(NotificationMessageHolderType::getMessage)
            .map(NotificationMessageHolderType.Message::getAny)
            .filter(any -> any instanceof NotifyContent)
            .map(any -> (NotifyContent) any)
            .forEach(n -> log.info(n.getIdType() + " : " + n.getId()));
    }

    @Test
    void test_message_replay() {
        pullPoint = nasClient.createPullPoint();
        subscribe = nasClient.subscribe(pullPoint, JOB_COMPLETED_TOPIC);

        JobStatusNotification jobStatusNotification = createJobStatusNotification();

        nasNotificationPublisher.publish(jobStatusNotification, "0506504003");
        nasNotificationPublisher.publish(jobStatusNotification, "0506504003");
        nasNotificationPublisher.publish(jobStatusNotification, "0506504003");
        nasNotificationPublisher.publish(jobStatusNotification, "0506504003");

        String messageId = UUID.randomUUID().toString();

        List<NotificationMessageHolderType> originalMessages = nasClient.getMessages(pullPoint, 1000, messageId);

        assertEquals(4, originalMessages.size());

        // replay request using same message id
        List<NotificationMessageHolderType> replayMessages = nasClient.getMessages(pullPoint, 1000, messageId);

        assertEquals(4, replayMessages.size());

        // use new message id
        List<NotificationMessageHolderType> newMessages = nasClient.getMessages(pullPoint, 1000, UUID.randomUUID().toString());

        assertEquals(0, newMessages.size());
    }

    private JobStatusNotification createJobStatusNotification() {
        ObjectFactory objectFactory = new ObjectFactory();
        JobStatusNotification jobStatusNotification = objectFactory.createJobStatusNotification();
        jobStatusNotification.setJobId(UUID.randomUUID().toString());
        jobStatusNotification.setJobStatus("OK");
        return jobStatusNotification;
    }

    private void setEndpoint(BindingProvider bindingProvider, String url) {
        bindingProvider.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, url);
    }
}
